Scriptname ReplaceThisWithYourModNamespace:HHSOutfit2 extends ObjectReference

; HHSOutfit Script - Version 2.6 rel
; Contributors: Tri0xin, niston, ...

Import HHS


; properties

Float Property HHSHeight Auto Const
{ The height offset to apply to actor when they equip this item (0 - 1000)}


; member vars

; initialization related
Int InitVersionTarget = 3 Const ; increment this if changes were made to initialization routine
Int InitVersionCurrent

; keyword array for default onSit exception keywords
Keyword[] OnSitExceptionKeywords

; internal variables for equipped status
Bool bIsEquipped
Actor actEquippedBy

; internal state for StopHeight/OnSit
Bool bIsStopped


; event handlers

Event OnInit()
	Initialize()
EndEvent

Event OnLoad()
	Initialize()
EndEvent

Event OnEquipped(Actor akActor)
	Debug.Trace(Self + " on Actor (" + akActor + "): Equipped by (" + akActor.GetLeveledActorBase().getName() + ") - Starting HHS.")
	StartHHS(akActor)	
EndEvent

Event OnUnequipped(Actor akActor)
	StopHHS(akActor)
	Debug.Trace(Self + " on Actor (" + akActor + "): Unequipped by (" + akActor.GetLeveledActorBase().getName() + ") - HHS stopped.")
EndEvent

Event Actor.OnDeath(Actor akActor, Actor akKiller)
	StopHHS(akActor)
	Debug.Trace(Self + " on Actor (" + akActor + "): Actor (" + akActor.GetLeveledActorBase().getName() + ") died - HHS stopped.")
EndEvent

Event actor.OnSit(Actor akActor, ObjectReference akFurniture)
	If (bIsEquipped)
		Debug.Trace(Self + " on Actor (" + akActor + "): Actor (" + akActor.GetLeveledActorBase().getName() + ") sat down - Updating Actor height.")
		UpdateHHSHeight(akActor)
	EndIf
EndEvent

Event actor.OnGetUp(Actor akActor, ObjectReference akFurniture)
	If (bIsEquipped)
		Debug.Trace(Self + " on Actor (" + akActor + "): Actor (" + akActor.GetLeveledActorBase().getName() + ") got up - Updating Actor height.")
		UpdateHHSHeight(akActor)
	EndIf
endEvent

Event ObjectReference.OnLoad(ObjectReference akRef)
	actor akActor = akRef as Actor
	ProcessOnLoad(akActor, "ObjectReference.OnLoad")
EndEvent

Event Actor.OnPlayerLoadGame(Actor akSender)
	actor akActor = GetContainer() as Actor
	Initialize()
	ProcessOnLoad(akActor, "Actor.OnPlayerLoadGame")
EndEvent

Event OnWorkshopObjectDestroyed(ObjectReference akActionRef)
	; we just got destroyed in workshop - so clean up
	EventsCleanup()	
	If ((bIsEquipped) && (actEquippedBy))
		; NOTE: does this actually ever happen? Idk...
		Debug.Trace(Self + " on Actor (" + actEquippedBy + "): Object scrapped/stored - Event registrations and internal references cleared.")
	Else
		Debug.Trace(Self +": Object scrapped/stored - Event registrations cleared.")
	EndIf
	actEquippedBy = none
	bIsEquipped = false	
EndEvent


; worker functions

Function UpdateHHSHeight(Actor akActor)	
	Int iIterations = 1
	Bool bExitLoop = false
	While ((!bExitLoop) && (iIterations < 3))
		If ((akActor as ObjectReference).Is3DLoaded())
			; check actor sleep/sit state
			If !(IsSitting(akActor) || IsSleeping(akActor))		
				If (bIsStopped)				
					StartHeight_HHS(akActor)
					bIsStopped = false
					Debug.Trace(Self + " on Actor (" + akActor + "): Height override restarted (not longer sitting or sleeping).")				
				EndIf
				; apply HHSHeight
				SetHeight_Internal(akActor, HHSHeight)
			Else
				; stop height
				StopHeight_HHS(akActor)
				bIsStopped = true
				Debug.Trace(Self + " on Actor (" + akActor + "): Height override stopped (sitting or sleeping).")
			EndIf
			bExitLoop = true
		Else
			If (iIterations == 1)
				; Actor has no 3D - Wait for it to load on first iteration
				akActor.WaitFor3DLoad()				
			Else
				; Waiting for Actor to load 3D was not successful. They will update when 3d is loaded later on.
				Debug.Trace(Self + ": UpdateHHSHeight(" + akActor + ") skipped - Actor has no 3D.")			
				bExitLoop = true
			EndIf
		EndIf
		iIterations += 1
	EndWhile	
EndFunction

; helper functions

Function StartHHS(Actor akActor)
	bIsEquipped = true	
	actEquippedBy = akActor
	EventsRegister(akActor)
	StartHeight_HHS(akActor)
	UpdateHHSHeight(akActor)
EndFunction

Function StopHHS(Actor akActor)
	EventsUnregister(akActor)
	RemoveHeight_HHS(akActor)
	bIsEquipped = false
	actEquippedBy = none
EndFunction

Function SetHeight_Internal(Actor akActor, Float HHSHeight)
	int retval = -99
	int tries = 0
	If ((akActor != none) && (HHSHeight >= 0))
		; try applying HHS height
		While (retval < 0) && (tries < 10)		
			retval = SetFloatHeight_HHS(akActor, HHSHeight)
			If (retval < 0)
				Debug.Trace(Self + ": SetFloatHeight_HHS(" + akActor + ") attempt failed; RetVal=" + retval + ". Retrying...")
				Utility.Wait(0.1)
			EndIf			
			tries += 1
		EndWhile			
		; failure message
		If (tries > 9)
			Debug.Trace(Self + " on Actor (" + akActor + "): Setting height failed for 10 out of 10 attempts. Giving up.")
		Else
			Debug.Trace(Self + " on Actor (" + akActor + "): Height set to " + HHSHeight + ".")
		EndIf								
	Else
		Debug.Trace(Self + ": SetHeight_Internal(" + akActor + ", " + HHSHeight + ") failed. Actor must not be none and HHSHeight must not have a negative value!")
	EndIf
EndFunction

Function ProcessOnLoad(Actor akActor, String eventName)
	If (akActor)		
		Debug.Trace(Self + " on " + akActor.GetLeveledActorBase().getName() + " (" + akActor + ") received " + eventName + " event.")
		
		If (bIsEquipped)
			If (akActor.Is3DLoaded())			
				Debug.Trace(Self + " on Actor (" + akActor + "): Item is equipped  - Starting HHS and updating Actor height.") 
				; start HHS
				StartHeight_HHS(akActor)
				; update height
				UpdateHHSHeight(akActor)		
			Else
				Debug.Trace(Self + " on Actor (" + akActor + "): Item is equipped but Actor has no 3D - Omitting HHS startup.") 
			EndIf
		Else
			Debug.Trace(Self + " on Actor (" + akActor + "): Item is not equipped - Omitting HHS startup.")			
		EndIf
	EndIf
EndFunction

Bool Function IsSitting(Actor akActor)	
	; actor sitting ?
	If (akActor.GetSitState() == 3)	
		; yes, obtain reference to furniture they sit in
		ObjectReference refFurniture = akActor.GetFurnitureReference()		
		If (refFurniture)
			; check furniture reference for OnSitException Keywords - now much faster and w/o any while loops :)			
			If (CommonArrayFunctions.CheckObjectAgainstKeywordArray(refFurniture, OnSitExceptionKeywords))
				Debug.Trace(Self + " on Actor (" + akActor + "): Furniture (" + refFurniture + ") has one or more OnSit Exception Keyword(s).")
				Return False
			EndIf				
		EndIf
		; no furniture or no exception keyword found
		Return True
	Else
		; actor is not sitting
		Return False
	EndIf
EndFunction

Bool Function IsSleeping(Actor akActor)
	Return (akActor.GetSleepState() == 3)
EndFunction

Function Initialize()
	; only initialize if new instance or if we say so (increment InitVersionTarget)	
	If (InitVersionCurrent < InitVersionTarget)

		; initialization routine

		; set up aray with default onsit exception keywords
		OnSitExceptionKeywords = new Keyword[18]
		OnSitExceptionKeywords[0] = Game.GetFormFromFile(0x024831, "Fallout4.esm") as Keyword ;AnimFurnWallLean
		OnSitExceptionKeywords[1] = Game.GetFormFromFile(0x183ad5, "Fallout4.esm") as Keyword ;AnimFurnClipboardWithPen
		OnSitExceptionKeywords[2] = Game.GetFormFromFile(0x0d96c6, "Fallout4.esm") as Keyword ;AnimFurnStandHammerHorizontal
		OnSitExceptionKeywords[3] = Game.GetFormFromFile(0x0d96c8, "Fallout4.esm") as Keyword ;AnimFurnStandHammerVertical
		OnSitExceptionKeywords[4] = Game.GetFormFromFile(0x05dd9c, "Fallout4.esm") as Keyword ;AnimFurnGuardPost
		OnSitExceptionKeywords[5] = Game.GetFormFromFile(0x137933, "Fallout4.esm") as Keyword ;AnimFurnStandingOnSoapBox
		OnSitExceptionKeywords[6] = Game.GetFormFromFile(0x0d96ca, "Fallout4.esm") as Keyword ;AnimFurnStandWrenchHorizontal
		OnSitExceptionKeywords[7] = Game.GetFormFromFile(0x0d96cc, "Fallout4.esm") as Keyword ;AnimFurnStandWrenchVertical
		OnSitExceptionKeywords[8] = Game.GetFormFromFile(0x16bae2, "Fallout4.esm") as Keyword ;AnimFurnNPCMapStand
		OnSitExceptionKeywords[9] = Game.GetFormFromFile(0x1b40bf, "Fallout4.esm") as Keyword ;AnimFurnNPCHandWarming
		OnSitExceptionKeywords[10] = Game.GetFormFromFile(0x1338f8, "Fallout4.esm") as Keyword ;AnimFurnNewspaperStanding
		OnSitExceptionKeywords[11] = Game.GetFormFromFile(0x1338f6, "Fallout4.esm") as Keyword ;AnimFurnNewspaperLeanRight
		OnSitExceptionKeywords[12] = Game.GetFormFromFile(0x1338f5, "Fallout4.esm") as Keyword ;AnimFurnNewspaperLeanLeft
		OnSitExceptionKeywords[13] = Game.GetFormFromFile(0x01a348, "Fallout4.esm") as Keyword ;AnimFurnLayOnGurneyLeftSide
		OnSitExceptionKeywords[14] = Game.GetFormFromFile(0x0f15d4, "Fallout4.esm") as Keyword ;AnimFurnLayDownInUtilityBox
		OnSitExceptionKeywords[15] = Game.GetFormFromFile(0x0c482f, "Fallout4.esm") as Keyword ;AnimFurnLayDown
		OnSitExceptionKeywords[16] = Game.GetFormFromFile(0x0fcb12, "Fallout4.esm") as Keyword ;AnimFurnDeskTerminalNoChair
		OnSitExceptionKeywords[17] = Game.GetFormFromFile(0x130bbb, "Fallout4.esm") as Keyword ;AnimFurnBerryGathering
		Debug.Trace(Self + ": OnSit Exceptions Keyword List updated; " + OnSitExceptionKeywords.Length + " keyword(s) configured on instance.")
	
		; update version number and write trace
		Int initVersionOld = InitVersionCurrent
		InitVersionCurrent = InitVersionTarget
		Debug.Trace(Self + ": Script instance initialized. InitVersionOld=" + initVersionOld + ", InitVersionCurrent=" + InitVersionCurrent + ".")
		
	EndIf
	
	If (IsBoundGameObjectAvailable())
		; ensure onload events are always registered if object is bound
		UnRegisterForRemoteEvent(Game.GetPlayer(), "OnPlayerLoadGame")
		RegisterForRemoteEvent(Game.GetPlayer(), "OnPlayerLoadGame")	
		If ((actEquippedBy != none) && (bIsEquipped))
			UnRegisterForRemoteEvent(actEquippedBy as ObjectReference, "OnLoad")
			RegisterForRemoteEvent(actEquippedBy as ObjectReference, "OnLoad")
		EndIf	
	EndIf
EndFunction

Function EventsRegister(Actor akActor)
	RegisterForRemoteEvent(akActor, "OnSit")
	RegisterForRemoteEvent(akActor, "OnGetUp")
	RegisterForRemoteEvent(akActor, "OnDeath")
	; re-register these, so we don't end up with multiple registrations (idk if it's a problem, but it sounds like one)...
	UnRegisterForRemoteEvent(akActor as ObjectReference, "OnLoad")
	RegisterForRemoteEvent(akActor as ObjectReference, "OnLoad")
	UnRegisterForRemoteEvent(Game.GetPlayer(), "OnPlayerLoadGame")
	RegisterForRemoteEvent(Game.GetPlayer(), "OnPlayerLoadGame")
EndFunction

Function EventsUnregister(Actor akActor)
	UnregisterForRemoteEvent(akActor, "OnSit")
	UnregisterForRemoteEvent(akActor, "OnGetUp")
	UnregisterForRemoteEvent(akActor, "OnDeath")
	UnregisterForRemoteEvent(akActor as ObjectReference, "OnLoad")
	UnRegisterForRemoteEvent(Game.GetPlayer(), "OnPlayerLoadGame")
EndFunction

Function EventsCleanup()
	If ((bIsEquipped) && (actEquippedBy))
		UnregisterForRemoteEvent(actEquippedBy, "OnSit")
		UnregisterForRemoteEvent(actEquippedBy, "OnGetUp")
		UnregisterForRemoteEvent(actEquippedBy, "OnDeath")
		UnregisterForRemoteEvent(actEquippedBy as ObjectReference, "OnLoad")
	EndIf
	UnRegisterForRemoteEvent(Game.GetPlayer(), "OnPlayerLoadGame")	
EndFunction